use ril::prelude::*;
use std::{fs::File, path::Path};

pub struct Template<T: AsRef<str>, V: AsRef<Path>> {
    elements: Vec<Element<T, V>>,
    dimensions: (u32, u32),
    base: Image,
}

impl<T: AsRef<str>, V: AsRef<Path>> Template<T, V> {
    pub fn new(base: Vec<u8>, elements: Vec<Element<T, V>>) -> Result<Self, CowError> {
        let base: Image<Dynamic> = Image::from_bytes_inferred(base)?;
        Ok(Self {
            dimensions: base.dimensions(),
            elements,
            base,
        })
    }

    pub fn produce(mut self) -> Vec<u8> {
        let size = self.base.dimensions();
        for element in self.elements {
            let mut pos = element.position;
            let mut img: Image = element.into();
            if pos.0 < 0 {
                img.crop(-pos.0 as u32, 0, size.0, size.1);
                pos.0 = 0;
            }
            if pos.1 < 0 {
                img.crop(0, -pos.1 as u32, size.0, size.1);
                pos.1 = 0;
            }

            self.base
                .draw(&Paste::new(img).with_position(pos.0 as u32, pos.1 as u32));
        }
        let mut buf = Vec::<u8>::new();
        self.base.encode(ImageFormat::Png, &mut buf).unwrap();

        buf
    }
}

pub struct Element<T: AsRef<str>, V: AsRef<Path>> {
    position: (i32, i32),
    dimensions: (u32, u32),
    media: Media<T, V>,
}

impl<T: AsRef<str>, V: AsRef<Path>> Element<T, V> {
    pub fn new(media: Media<T, V>, position: (i32, i32), dimensions: (u32, u32)) -> Self {
        Self {
            position,
            dimensions,
            media,
        }
    }
}

impl<T: AsRef<str>, V: AsRef<Path>> Into<Image> for Element<T, V> {
    fn into(self) -> Image {
        match self.media {
            Media::Image(image) => image.resized(
                self.dimensions.0,
                self.dimensions.1,
                ResizeAlgorithm::Bicubic,
            ),
            Media::Text(text) => Image::new(
                self.dimensions.0,
                self.dimensions.1,
                Dynamic::Rgba(Rgba::transparent()),
            )
            .with(
                &TextSegment::new(
                    &Font::from_reader(File::open(text.font).unwrap(), 12.0).unwrap(),
                    text.text,
                    Dynamic::Rgb(Rgb::new(text.fill.0, text.fill.1, text.fill.2)),
                )
                .with_size(text.size),
            ),
        }
    }
}

pub enum Media<T: AsRef<str>, V: AsRef<Path>> {
    Text(Text<T, V>),
    Image(Image),
}

pub struct Text<T: AsRef<str>, V: AsRef<Path>> {
    text: T,
    font: V,
    size: f32,
    fill: (u8, u8, u8),
}

impl<T: AsRef<str>, V: AsRef<Path>> Text<T, V> {
    pub fn new(text: T, font: V, size: f32, fill: (u8, u8, u8)) -> Self {
        Self {
            text,
            font,
            size,
            fill,
        }
    }
}

#[derive(Clone, Debug)]
pub enum CowError {
    Other(String),
}

impl From<ril::Error> for CowError {
    fn from(r: ril::Error) -> Self {
        CowError::Other(r.to_string())
    }
}

#[cfg(test)]
mod tests {
    #[test]
    fn it_works() {
        let result = 2 + 2;
        assert_eq!(result, 4);
    }
}
